'''Database abstraction for key/value stores.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

Many popular large-scale databases export a simple key/value
abstraction: the database is simply a list of cells with some
(possibly structured) key and a value for each key.  This allows the
database system itself to partition the database in some way, and in a
distributed system it allows the correct system hosting a specific key
to be found easily.  This model is also simple enough that it can be
used with in-memory storage or more traditional SQL-based databases.

This module provides a simple abstraction around these
key/value-oriented databases.  It works with :mod:`yakonfig` to hold
basic configuration settings, so the top-level YAML configuration must
have a ``kvlayer`` block.  This will typically look something like

.. code-block:: yaml

    kvlayer:
      storage_type: redis
      storage_addresses: [redis.example.com:6379]
      app_name: app
      namespace: namespace

These four parameters are always required.  ``storage_type`` gives one
of the database backends described below.  ``storage_addresses`` is a
list of backend-specific database locations.  ``app_name`` and
``namespace`` combine to form a container for the virtual tables
stored in kvlayer.

Backend-specific configuration may also be broken into a separate
section, whose name matches the ``storage_type``:

.. code-block:: yaml

    kvlayer:
      storage_type: redis
      redis:
        storage_addresses: [redis.example.com:6379]
        app_name: app
        namespace: namespace

This is useful if experimenting with different backends, or with the
``split_s3`` hybrid backend.  Any configuration value may be in either
the backend-specific configuration or the top-level kvlayer
configuration, with deeper values taking precedence.

Backends
========

local
-----

This is intended only for testing.  Values are stored in a Python
dictionary and not persisted.  ``storage_addresses`` are not required.

.. code-block:: yaml

    kvlayer:
      storage_type: local

filestorage
-----------

This is intended only for testing.  Values are stored in a local file
using the :mod:`shelve` module.  This does not require
``storage_addresses``, but it does require:

.. code-block:: yaml

    kvlayer:
      storage_type: filestorage

      # Name of the file to use for storage
      filename: /tmp/kvlayer.bin

      # If set, actually work on a copy of "filename" at this location.
      copy_to_filename: /tmp/kvlayer-copy.bin

redis
-----

Uses the `Redis`_ in-memory database.

.. code-block:: yaml

    kvlayer:
      storage_type: redis

      # host:port locations of Redis servers; only the first is used
      storage_addresses: [redis.example.com:6379]

      # Redis database number (default: 0)
      redis_db_num: 1

accumulo
--------

Uses the Apache `Accumulo`_ distributed database.  Your installation
must be running the Accumulo proxy, and the configuration points at
that proxy.

.. code-block:: yaml

    kvlayer:
      storage_type: accumulo

      # host:port location of the proxy; only the first is used
      storage_addresses: [accumulo.example.com:50096]
      username: root
      password: secret

      # all of the following parameters are default values and are optional
      accumulo_max_memory: 1000000
      accumulo_timeout_ms: 30000
      accumulo_threads: 10
      accumulo_latency_ms: 10
      thrift_framed_transport_size_in_mb: 15

Each kvlayer table is instantiated as an Accumulo table named
``appname_namespace_table``.

.. _Accumulo: http://accumulo.apache.org/

postgrest
---------

Uses `PostgreSQL`_ for storage.  This backend is only available if the
:mod:`psycopg2` module is installed.  The ``storage_addresses`` may be
a single `PostgreSQL connection string`_, or may be a ``host:port``
format with additional configuration.  The ``app_name`` and
``namespace`` can only consist of alphanumeric characters,
underscores, or ``$``, and must begin with a letter or underscore.

This backend is newer than, less proven than, and incompatible with the
``postgres`` backend described below.

.. code-block:: yaml

    kvlayer:
      storage_type: postgrest
      storage_addresses: ['postgres.example.com:5432']
      username: test
      password: test
      dbname: test
      # Equivalently, pack this all into a single SQL connection string
      # storage_addresses:
      # - 'host=postgres.example.com port=5432 user=test dbname=test password=test'

      # all of the following parameters are default values and are optional
      # keep this many connections alive
      min_connections: 2
      # never create more than this many connections
      max_connections: 16

The backend assumes the user is able to run SQL ``CREATE TABLE`` and
``DROP TABLE`` statements.  Each kvlayer table is instantiated as an
SQL table named ``appname_namespace_table``.

Within the system, the ``min_connections`` and ``max_connections``
property apply per client object.  If ``min_connections`` is set to 0
then the connection pool will never hold a connection alive, which
typically adds a performance cost to reconnect.

postgres
--------

Uses `PostgreSQL`_ for storage.  This backend is only available if the
:mod:`psycopg2` module is installed.  The ``storage_type`` is a
`PostgreSQL connection string`_.  The ``app_name`` and ``namespace``
can only consist of alphanumeric characters, underscores, or ``$``,
and must begin with a letter or underscore.

.. code-block:: yaml

    kvlayer:
      storage_type: postgres
      storage_addresses:
      - 'host=postgres.example.com port=5432 user=test dbname=test password=test'

      # all of the following parameters are default values and are optional
      # keep this many connections alive
      min_connections: 2
      # never create more than this many connections
      max_connections: 16
      # break large scans (using SQL) into chunks of this many
      scan_inner_limit: 1000

The backend assumes the user is able to run SQL ``CREATE TABLE`` and
``DROP TABLE`` statements.  Each kvlayer namespace is instantiated as
an SQL table named ``kv_appname_namespace``; kvlayer tables are
collections of rows within the namespace table sharing a common field.

Within the system, the ``min_connections`` and ``max_connections``
property apply per client object.  If ``min_connections`` is set to 0
then the connection pool will never hold a connection alive, which
typically adds a performance cost to reconnect.

.. _PostgreSQL: http://www.postgresql.org
.. _PostgreSQL connection string: http://www.postgresql.org/docs/current/static/libpq-connect.html#LIBPQ-PARAMKEYWORDS

riak
----

Uses `Riak`_ for storage.  This backend is only available if the
corresponding :mod:`riak` client library is installed.  Multiple
``storage_addresses`` are actively encouraged for this backend. Each
may be a simple string, or a dictionary containing keys ``host``,
``http_port``, and ``pb_port`` if your setup is using non-standard
port numbers.  A typical setup will look like:

.. code-block:: yaml

    kvlayer:
      storage_type: riak
      storage_addresses: [riak01, riak02, riak03, riak04, riak05]
      # optional settings with their default values
      protocol: pbc # or http or https
      scan_limit: 100

The setup from the Riak "Five-Minute Install" runs five separate
Riak nodes all on localhost, resulting in configuration like

.. code-block: yaml

    kvlayer:
      storage_type: riak
      storage_addresses:
      - { host: "127.0.0.1", pb_port: 10017, http_port: 10018 }
      - { host: "127.0.0.1", pb_port: 10027, http_port: 10028 }
      - { host: "127.0.0.1", pb_port: 10037, http_port: 10038 }
      - { host: "127.0.0.1", pb_port: 10047, http_port: 10048 }
      - { host: "127.0.0.1", pb_port: 10057, http_port: 10058 }

One :mod:`kvlayer` namespace corresponds to one Riak bucket.

The ``protocol`` setting selects between defaulting to the HTTP or
protocol buffer APIs.  While Riak's default is generally ``http``,
the ``pbc`` API seems to work equally well and is much faster.  The
kvlayer backend's default is ``pbc``.

The ``scan_limit`` setting determines how many results will be
returned from each secondary index search.  A higher setting for this
results in fewer network round-trips to get search results, but also
results in higher latency to return each.  This affects both calls to
the kvlayer scan API as well as calls to delete kvlayer tables, which
are also Riak key scans.

Your Riak cluster must be configured with secondary indexing enabled,
and correspondingly, must be using the LevelDB backend.  The default
bucket settings, and in particular setting ``allow_mult`` to
``false``, are correct for :mod:`kvlayer`.

split\_s3
---------

Hybrid backend that stores values for specific tables in `Amazon S3`_,
and pointers to those values and all other tables in some other
:mod:`kvlayer` backend.  This backend is only available if the
:mod:`boto` AWS access library is installed.  Using this effectively
requires knowing the internal details of what table names the end
application will use.  A typical all-Amazon setup would store bulk
objects in S3, and store all other objects in an `Amazon RDS`_
PostgreSQL instance.

.. code-block:: yaml

    kvlayer:
      storage_type: split_s3
      app_name: kvlayer
      namespace: namespace
      split_s3:
        tables: [stream_items]
        aws_access_key_id: Axxxxxxxxxxxxxxxxxxx
        aws_secret_access_key: a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0a0
        bucket: mybucket
        path_prefix:
        kvlayer_prefix: True
        retries: 5
        retry_interval: 0.1
        kvlayer:
          storage_type: postgres
          storage_addresses:
            - >-
                host=stuff.rds.amazonaws.com port=5432 user=db
                dbname=db password=db

An AWS key must be passed to the backend.  Either an access key ID and
the secret key can be embedded directly in the configuration, as shown
here, or ``aws_access_key_id_path`` and ``aws_secret_access_key_path``
can be names of local files containing those credentials.

Objects are stored in the named ``bucket``, which must exist already.
Each object's key is constructed by taking ``path_prefix`` (if any);
if ``kvlayer_prefix`` is true (default), appending
``app_name/namespace/``; ``table_name/``; and a SHA-256 hash of the
serialized object key, with the first two bytes split off to form a
directory hierarchy.  If your table's schema is ``(str,)`` and you
wrote keys ``a``, ``b``, and ``c`` with the above configuration to
table ``table``, these would be written to ``mybucket`` in respective
S3 object paths

    kvlayer/namespace/table/ca/97/8112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb
    kvlayer/namespace/table/3e/23/e8160039594a33894f6564e1b1348bbd7a0088d42c4acb73eeaed59c009d
    kvlayer/namespace/table/2e/7d/2c03a9507ae265ecf5b5356885a53393a2029d241394997265a1a25aefc6

The kvlayer keys are written through to the underlying backend, but no
data is written there.  Only tables listed in the ``tables`` parameter
have data stored in S3 (required setting); all other tables are
processed normally.
:meth:`~kvlayer._abstract_storage.AbstractStorage.scan_keys` does not
talk to S3 at all, it just relays the scan from the underlying
backend.  Bulk-delete operations such as
:meth:`~kvlayer._abstract_storage.AbstractStorage.clear_table` and
:meth:`~kvlayer._abstract_storage.AbstractStorage.delete_namespace`
need to do a scan of the underlying table to find the keys to delete.
Particularly if data is being read back immediately after being
written, it is possible that a read (get or scan) operation will fail
to find the just-written object; a failed read for an object expected
to exist will be retried ``retries`` times (default 5), waiting
``retry_interval`` seconds (default 0.1) between each.

The underlying backend is separately configured with its own
``kvlayer`` section inside the backend configuration.

.. _Amazon S3: https://aws.amazon.com/s3/
.. _Amazon RDS: https://aws.amazon.com/rds/

cassandra
---------

Uses the Apache `Cassandra`_ distributed database.  Note that this
backend requires keys to be limited to tuples of UUIDs.

.. code-block:: yaml

    kvlayer:
      storage_type: cassandra
      storage_addresses: ['cassandra.example.com:9160']
      username: root
      password: secret

      connection_pool_size: 2
      max_consistency_delay: 120
      replication_factor: 1
      thrift_framed_transport_size_in_mb: 15

.. _Cassandra: http://cassandra.apache.org/

API
===

Having set up the global configuration, it is enough to call
:func:`kvlayer.client` to get a storage client object.

The API works in terms of "tables", though these are slightly
different from tradational database tables.  Each table has keys which
are tuples of a fixed length.

.. autofunction:: client

.. autoclass:: kvlayer._abstract_storage.AbstractStorage
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: DatabaseEmpty
.. autoclass:: BadKey

Instance Collections
====================

.. automodule:: kvlayer.instance_collection

'''

from kvlayer._client import client
from kvlayer.config import config_name, default_config, add_arguments, \
    runtime_keys, check_config
from kvlayer._exceptions import DatabaseEmpty, BadKey



'''Table-per-table kvlayer PostgreSQL backend.

.. This software is released under an MIT/X11 open source license.
   Copyright 2014 Diffeo, Inc.

At a high level this is very similar to :mod:`kvlayer._postgres`.
However, that implementation puts all kvlayer data into a single SQL
table with a single string key.  If there is a large amount of data
this table can become very large, and the unstructured string keys
limit the database's abilities to do some kinds of searches.

Since the underlying SQL tables look totally different, this isn't
compatible with the older implementation, and since all of the queries
are totally different, it makes more sense to just do a separate
implementation.

'''
# Also, there's a lot of overlap with kvlayer._postgres.  I expect that
# this will *replace* that implementation and maybe become a base for
# a future generic-SQL backend, so I'm not going to be embarrassed by
# the copy-and-paste.

from __future__ import absolute_import
import contextlib
import logging
import re
import uuid

import psycopg2
import psycopg2.errorcodes
import psycopg2.extras
import psycopg2.pool

from kvlayer._abstract_storage import AbstractStorage
from kvlayer._exceptions import ConfigurationError, ProgrammerError

logger = logging.getLogger(__name__)
psycopg2.extras.register_uuid()

class PostgresTableStorage(AbstractStorage):
    '''PostgreSQL kvlayer backend.'''
    config_name = 'postgrest'
    default_config = {
        'min_connections': 2,
        'max_connections': 16,
    }

    def __init__(self, *args, **kwargs):
        '''Create a new PostgreSQL kvlayer client.'''
        super(PostgresTableStorage, self).__init__(*args, **kwargs)
        
        # Is the namespace string valid?
        # http://www.postgresql.org/docs/9.3/static/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS
        # has the real rules.
        if not re.match('[a-z_][a-z0-9_$]*', self._app_name, re.IGNORECASE):
            raise ConfigurationError('app_name {!r} must be a valid SQL name'
                                     .format(self._app_name))
        if not re.match('[a-z_][a-z0-9_$]*', self._namespace, re.IGNORECASE):
            raise ConfigurationError('namespace {!r} must be a valid SQL name'
                                     .format(self._namespace))

        # Figure out what we're connecting to
        storage_addresses = self._config.get('storage_addresses', [])
        if not storage_addresses:
            raise ConfigurationError('no storage_addresses for postgrest')
        if len(storage_addresses) > 1:
            logger.warning('multiple storage_addresses for postgrest, '
                           'using first only')
        connect_string = storage_addresses[0]
        # This is cheap, but a multi-part k=v string and a host[:port]
        # string should be pretty easy to tell apart
        if '=' not in connect_string and ' ' not in connect_string:
            if ':' in connect_string:
                (host, port) = connect_string.split(':', 2)
                connect_string = 'host={} port={}'.format(host, port)
            else:
                connect_string = 'host={}'.format(connect_string)
            user = self._config.get('username', None)
            if not user:
                raise ConfigurationError('no username for postgrest')
            password = self._config.get('password', None)
            if not password:
                raise ConfigurationError('no password for postgrest')
            dbname = self._config.get('dbname', None)
            if not dbname:
                raise ConfigurationError('no dbname for postgrest')
            connect_string += ' user={} password={} dbname={}'.format(
                user, password, dbname)
        self.connection_pool = psycopg2.pool.SimpleConnectionPool(
            self._config.get('min_connections', 2),
            self._config.get('max_connections', 16),
            connect_string
        )

    @contextlib.contextmanager
    def _conn(self):
        '''Produce a PostgreSQL connection from the pool.

        This also runs a single transaction on that connection.  On
        successful completion, the transaction is committed; if any
        exception is thrown, the transaction is aborted.

        On successful completion the connection is returned to the
        pool for reuse.  If any exception is thrown, the connection
        is closed.

        '''
        conn = self.connection_pool.getconn()
        try:
            with conn:
                yield conn
        finally:
            # This has logic to test whether the connection is closed
            # and/or failed and correctly manages returning it to the
            # pool (or not).
            self.connection_pool.putconn(conn)

    @contextlib.contextmanager
    def _cursor(self, name=None):
        '''Produce a cursor from a connection.

        This is a helper for the common case of wanting a single
        cursor on a single connection to do a single operation.

        '''
        with self._conn() as conn:
            with conn.cursor(name=name) as cursor:
                yield cursor

    def _table_name(self, table_name):
        '''Get the SQL table name of a kvlayer table.'''
        return '{}_{}_{}'.format(self._app_name, self._namespace, table_name)

    def _columns(self, key_spec):
        '''Get the names of the columns for a specific table.'''
        return ['k{}'.format(n+1) for n in xrange(len(key_spec))]

    def _massage_key_part(self, typ, kp):
        '''Change a single key part if needed.'''
        if typ == str:
            return psycopg2.Binary(kp)
        return kp

    def _massage_key_tuple(self, key_spec, key):
        '''Change types in `key` to things that can be passed into postgres.'''
        return tuple(self._massage_key_part(typ, kp)
                     for typ, kp in zip(key_spec, key))

    def _massage_result_part(self, typ, rp):
        '''Change a single result part if needed.'''
        if typ == str:
            return rp[:]
        return rp

    def _massage_result_tuple(self, key_spec, row):
        '''Change types in `row` from postgres return types.'''
        return tuple(self._massage_result_part(typ, rp)
                     for typ, rp in zip(key_spec, row))

    def _python_to_sql_type(self, typ):
        if typ is int:
            return 'INTEGER'
        if typ is long:
            # We'd think "BIGINT" (64-bit) would be enough for
            # practical uses, but Python long is unbounded.
            return 'NUMERIC(1000,0)'
        if typ is uuid.UUID:
            return 'UUID'
        if typ is str: # but not unicode; bytes in Python 3
            return 'BYTEA'
        if isinstance(typ, tuple):
            # Other backends just pass typ as the second argument to
            # isinstance() and don't use the information to serialize.
            # Right now we can handle this iff the types are all
            # "similar", particularly integers.
            typs = [(long, int)]
            for candidates in typs:
                if all(t in candidates for t in typ):
                    return self._python_to_sql_type(candidates[0])
        raise ProgrammerError('unexpected key type {!r}'.format(typ))

    def setup_namespace(self, table_names):
        '''Create tables in the namespace.

        One table is created in PostgreSQL for each table in
        `table_names`.  As elsewhere in :mod:`kvlayer`, if a table's
        value is an integer, it is interpreted as a tuple of that many
        :class:`uuid.UUID` objects.  Otherwise, this accepts tuples
        of :func:`int`, :func:`long`, :class:`uuid.UUID`, and
        :func:`str`; these are mapped to the SQL types ``INTEGER``,
        ``BIGINT``, ``UUID``, and ``BYTEA``, repsectively.  (The last
        two are PostgreSQL extensions.)

        '''
        super(PostgresTableStorage, self).setup_namespace(table_names)
        for name, key_spec in self._table_names.iteritems():
            with self._cursor() as cursor:
                cnames = self._columns(key_spec)
                ctypes = [self._python_to_sql_type(t) for t in key_spec]
                columns = ['{} {} NOT NULL'.format(n, t)
                           for n, t in zip(cnames, ctypes)]
                sql = ('CREATE TABLE IF NOT EXISTS {} ({}, v BYTEA NOT NULL, '
                       'PRIMARY KEY ({}))'
                       .format(self._table_name(name), ', '.join(columns),
                               ', '.join(cnames)))
                cursor.execute(sql)

                # Create an "upsert" stored procedure.
                # There's a reference PostgreSQL one at
                # http://www.postgresql.org/docs/9.3/static/plpgsql-control-structures.html
                # and also some commentary at
                # http://johtopg.blogspot.com/2014/04/upsertisms-in-postgres.html
                # There are three obvious approaches:
                # (1) SELECT for the row, and INSERT or UPDATE as appropriate
                # (2) INSERT the row, UPDATEing on exception
                # (3) Do a single-statement INSERT...WHERE NOT EXISTS
                #     (e.g., http://www.the-art-of-web.com/sql/upsert/)
                # The second link suggests that (3) is never fastest
                # (but typically never slowest either), and that (1) is
                # faster if the row usually exists (>30-40%) and (2)
                # is faster otherwise.  Commentary further notes, the "try"
                # isn't expensive, the "catch" is.
                #
                # Assume we're usually inserting, not updating.
                # (May consider making this a knob.)
                
                # Function parameters
                args = ['IN {}p {}'.format(n, t)
                        for n, t in zip(cnames, ctypes)]
                sql = '''
                CREATE FUNCTION upsert_{0}({1}, IN vp BYTEA)
                RETURNS VOID AS
                $$
                BEGIN
                  LOOP
                    BEGIN
                      INSERT INTO {0}({2}, v) VALUES ({3}, vp);
                      RETURN;
                    EXCEPTION WHEN unique_violation THEN
                      UPDATE {0} SET v=vp WHERE {4};
                      IF found THEN
                        RETURN;
                      END IF;
                    END;
                  END LOOP;
                END;
                $$
                LANGUAGE plpgsql
                '''.format(self._table_name(name),
                           ', '.join(args), 
                           ', '.join(cnames),
                           ', '.join(k + 'p' for k in cnames),
                           ' AND '.join('{0}={0}p'.format(k) for k in cnames))
                try:
                    cursor.execute(sql)
                except psycopg2.ProgrammingError, e:
                    # Ignore "function already exists"
                    if e.pgcode != psycopg2.errorcodes.DUPLICATE_FUNCTION:
                        raise
                    # If we're here, though, also note that the transaction
                    # is aborted...this is okay since we generally assume
                    # the table and stored procedure go together,
                    # and the outer loop uses a transaction per table
                    pass

    def delete_namespace(self):
        '''Delete all of the known tables.'''
        with self._cursor() as cursor:
            for name, key_spec in self._table_names.iteritems():
                tn = self._table_name(name)
                cnames = self._columns(key_spec)
                ctypes = [self._python_to_sql_type(t) for t in key_spec]
                cursor.execute('DROP TABLE {}'.format(tn))
                args = ['IN {} {}'.format(n, t)
                        for n, t in zip(cnames, ctypes)]
                cursor.execute('DROP FUNCTION upsert_{}({}, IN v BYTEA)'
                               .format(tn, ', '.join(args)))

    def clear_table(self, table_name):
        '''Clear out a single table.'''
        with self._cursor() as cursor:
            cursor.execute('TRUNCATE {}'.format(self._table_name(table_name)))

    def put(self, table_name, *keys_and_values, **kwargs):
        '''Write data into a table.'''
        key_spec = self._table_names[table_name]
        tn = self._table_name(table_name)
        with self._cursor() as cursor:
            for k, v in keys_and_values:
                self.check_put_key_value(k, v, table_name, key_spec)
                k = self._massage_key_tuple(key_spec, k)
                v = self._massage_key_part(str, v)
                cursor.callproc('upsert_' + tn, k + (v,))

    def get(self, table_name, *keys, **kwargs):
        '''Get values out of the database.'''
        tn = self._table_name(table_name)
        key_spec = self._table_names[table_name]
        cnames = self._columns(key_spec)
        exprs = ['{}=%s'.format(kn) for kn in cnames]
        where = 'WHERE {}'.format(' AND '.join(exprs))
        sql = 'SELECT v FROM ' + tn + ' ' + where
        with self._cursor() as cursor:
            for k in keys:
                k = self._massage_key_tuple(key_spec, k)
                cursor.execute(sql, k)
                found = False
                for row in cursor:
                    assert not found
                    found = True
                    yield k, self._massage_result_part(str, row[0])
                if not found:
                    yield k, None

    def _scan_padded(self, key_spec, k):
        '''Add :const:`None` to the end of `k` so it's the right length'''
        if k is None:
            k = ()
        k += (None,) * (len(k) - len(key_spec))
        return k

    def _scan_where_one(self, cnames, key_spec, op, key):
        '''``tuple(kN) `op` key`` as part of a WHERE clause'''
        z = zip(cnames, self._massage_key_tuple(key_spec, key))
        query = ' AND '.join(['{}{}%s'.format(cn, op) for (cn, kp) in z])
        wt = tuple(kp for (cn, kp) in z)
        return (query, wt)

    def _scan_where_parts(self, cnames, key_spec, lo, hi):
        '''``tuple(kN) >= pair[0] AND tuple(kN) <= pair[1]``'''
        # Simple, expected cases:
        # (None,(anything)) ==> tuple(kN) <= anything
        # ((anything),None) ==> tuple(kN) >= anything
        # ((k1val,),(kp1val,)) ==> k1 == k1val (ignoring k2, k3, ...)
        # ((k1lo,),(k1hi,)) ==> k1 >= k1lo AND k1 <= k1hi
        # ((k1val,k2lo),(k1val,)) ==> k1=k1val and k2>=k2lo
        # ((k1val,k2lo),(k1val,k2hi)) ==> k1=k1val AND k2>=k2lo AND k2<=k2hi
        #
        # Harder cases:
        # ((a,b),(c,d)) breaks down into
        #  k1=a and k2>=b
        #  k1>a and k1<c
        #  k1=c and k2<=d
        # ((a,b),(c,) breaks down into
        #  k1=a and k2>=b
        #  k1>a and k1<=c
        if len(lo) == 0 and len(hi) == 0:
            return ('', ())
        if len(hi) == 0:
            return self._scan_where_one(cnames, key_spec, '>=', lo)
        if len(lo) == 0:
            return self._scan_where_one(cnames, key_spec, '<=', hi)
        if lo[0] == hi[0]:
            (query, params) = self._scan_where_parts(
                cnames[1:], key_spec[1:], lo[1:], hi[1:])
            query = cnames[0] + '=%s' + (query and ' AND ' + query)
            return (query,
                    (self._massage_key_part(key_spec[0], lo[0]),) + params)
        assert lo[0] < hi[0]
        if len(lo) == 1 and len(hi) == 1:
            return (cnames[0] + '>=%s AND ' + cnames[0] + '<=%s',
                    (self._massage_key_part(key_spec[0], lo[0]),
                     self._massage_key_part(key_spec[0], hi[0])))
        # Otherwise: the two keys differ in their first part, and
        # this isn't a trivial scan.
        if len(lo) == 1:
            # (a,) to (c,d)
            # So, (k1>=a AND k1<c) OR (k1=c and k2<d)
            #     (a,) to (<c,)    OR (c,) to (c,d)
            (query, params) = self._scan_where_parts(
                cnames, key_spec, hi[:1], hi)
            my_query = cnames[0] + '>=%s AND ' + cnames[0] + '<%s'
            if query:
                query = '(' + my_query + ') OR (' + query + ')'
            else:
                query = my_query
            params = (self._massage_key_part(key_spec[0], lo[0]),
                      self._massage_key_part(key_spec[0], hi[0])) + params
            return (query, params)
        if len(hi) == 1:
            # (a,b) to (c,)
            # So, (k1=a AND k2>=b) OR (k1>a AND k1<=c)
            #     (a,b) to (a,)    OR (>a,) to (c,)
            (query, params) = self._scan_where_parts(
                cnames, key_spec, lo, lo[:1])
            my_query = cnames[0] + '>%s AND ' + cnames[0] + '<=%s'
            if query:
                query = '(' + query + ') OR (' + my_query + ')'
            else:
                query = my_query
            params = params + (self._massage_key_part(key_spec[0], lo[0]),
                               self._massage_key_part(key_spec[0], hi[0]))
            return (query, params)
        # (a,b) to (c,d)
        # This is a combination of all of the above:
        # (k1=a AND k2>=b) OR (k1>a AND k1<c) OR (k1=c AND k2<=d)
        (q1, p1) = self._scan_where_parts(
            cnames, key_spec, lo, lo[:1]) # (a,b) to (a,)
        q2 = cnames[0] + '>%s AND ' + cnames[0] + '<%s'
        p2 = (self._massage_key_part(key_spec[0], lo[0]),
              self._massage_key_part(key_spec[0], hi[0]))
        (q3, p3) = self._scan_where_parts(
            cnames, key_spec, hi[:1], hi) # (c,) to (c,d)
        if q1:
            if q3:
                query = '({}) OR ({}) OR ({})'.format(q1, q2, q3)
            else:
                query = '({}) OR ({})'.format(q1, q2)
        else:
            if q3:
                query = '({}) OR ({})'.format(q2, q3)
            else:
                query = q2
        params = p1 + p2 + p3
        return (query, params)

    def _scan_where(self, cnames, key_spec, key_ranges):
        '''Build an SQL WHERE clause for a scan.

        Returns a pair of the actual WHERE clause and the values that need
        to be passed into it.'''
        if len(key_ranges) == 0 or (None, None) in key_ranges:
            return '', () # just scan the whole table
        parts = [self._scan_where_parts(cnames, key_spec,
                                        list(lo or []), list(hi or []))
                 for (lo, hi) in key_ranges]
        parts = [(query, params) for (query, params) in parts if query]
        where = ''
        if parts:
            where = ' OR '.join(['({})'.format(p[0]) for p in parts])
            where = 'WHERE ' + where
        wt = sum([p[1] for p in parts], ())
        return (where, wt)

    def scan(self, table_name, *key_ranges, **kwargs):
        '''Get ordered (key, value) pairs out of the database.'''
        # We have previously asserted that there is a significant cost
        # to running a very large scan in a transaction attributable to
        # keeping corresponding row versions around.  See the
        # scan_inner_limit parameter in _postgres.py and its
        # implementation.
        tn = self._table_name(table_name)
        key_spec = self._table_names[table_name]
        cnames = self._columns(key_spec)
        (where, wt) = self._scan_where(cnames, key_spec, key_ranges)
        order = 'ORDER BY {}'.format(', '.join(cnames))
        with self._cursor(name='scan') as cursor:
            sql = ('SELECT {}, v FROM {} {} {}'
                   .format(', '.join(cnames), tn, where, order))
            cursor.execute(sql, wt)
            for row in cursor:
                k = row[:-1]
                v = row[-1]
                yield (self._massage_result_tuple(key_spec, k),
                       self._massage_result_part(str, v))

    def scan_keys(self, table_name, *key_ranges, **kwargs):
        '''Get ordered key tuples out of the database.'''
        # Just like scan().
        tn = self._table_name(table_name)
        key_spec = self._table_names[table_name]
        cnames = self._columns(key_spec)
        (where, wt) = self._scan_where(cnames, key_spec, key_ranges)
        order = 'ORDER BY {}'.format(', '.join(cnames))
        with self._cursor(name='scan') as cursor:
            cursor.execute('SELECT {} FROM {} {} {}'
                           .format(', '.join(cnames), tn, where, order),
                           wt)
            for row in cursor:
                yield self._massage_result_tuple(key_spec, row)

    def delete(self, table_name, *keys, **kwargs):
        tn = self._table_name(table_name)
        key_spec = self._table_names[table_name]
        cnames = self._columns(key_spec)
        exprs = ['{}=%s'.format(kn) for kn in cnames]
        where = 'WHERE {}'.format(' AND '.join(exprs))
        sql = 'DELETE FROM ' + tn + ' ' + where
        with self._cursor() as cursor:
            for k in keys:
                k = self._massage_key_tuple(key_spec, k)
                cursor.execute(sql, k)

    def close(self):
        '''
        close connections and end use of this storage client
        '''
        if self.connection_pool:
            try:
                self.connection_pool.closeall()
            finally:
                self.connection_pool = None



import pytest
import logging

logger = logging.getLogger(__name__)

config_postgres = {
    'namespace': 'test',
    'storage_addresses': ['host=test-postgres.diffeo.com port=5432 user=test dbname=test password=test'],
}

try:
    from kvlayer._postgres import PGStorage, _valid_namespace
    postgres_missing = 'False'
except ImportError:
    postgres_missing = 'True'

@pytest.mark.skipif(postgres_missing)
@pytest.mark.parametrize(
    "badnamespace",
    [None,
     '9isnotaletter',
     '$isnotaletter',
 ])
def test_illegal_namespaces(badnamespace):
    with pytest.raises(Exception):
        config = dict(config_postgres)
        config['namespace'] = badnamespace
        pg = PGStorage(config)


@pytest.mark.skipif(postgres_missing)
@pytest.mark.parametrize(
    "namespace",
    ['_ok', 'Aok', 'aok'])
def test_legal_namespaces(namespace):
    assert _valid_namespace(namespace)



from __future__ import division, absolute_import
import argparse
import cProfile
import multiprocessing
import logging
import os
import Queue
from signal import alarm, signal, SIGHUP, SIGTERM, SIGABRT, SIGALRM
import sys
import uuid
import time

import pytest

import kvlayer
from kvlayer.tests.test_interface import client, backend # fixture
import yakonfig

logger = logging.getLogger(__name__)

def worker(operation, i_queue, o_queue, tasks_remaining,
           class_config=None, profile=False
    ):
    '''simple worker used by run_many below

@operation(task, o_queue): callable that takes "task" received from
i_queue and optionally passes something to o_queue

@i_queue: a multiprocessing Queue provided by run_many
@o_queue: a multiprocessing Queue provided by run_many
@tasks_remaining: a multiprocessing Event provided by run_many

@class_config: if not None, then operation is initialized by calling
operation(class_config) before using it inside each worker.

    '''
    pr = None
    if profile:
        pr = cProfile.Profile()
        pr.enable()
    if class_config is not None:
        ## operation is a class with __call__ so initialize it
        operation = operation(class_config)
    while 1:
        try:
            task = i_queue.get(timeout=0.3)
            operation(task, o_queue)
        except Queue.Empty:
            if tasks_remaining.is_set():
                continue
            else:
                break
    if pr:
        pr.create_stats()
        pr.dump_stats('profile.%d.txt' % os.getpid())

def run_many(operation, task_generator, num_workers=1, timeout=None,
             class_config=None, profile=False,
    ):
    '''test harness that runs num_workers as child processes and sends
them tasks from task_generator via queue.  Results are harvested from
o_queue and yielded by this function, so typical usage is:

for response from run_many(do_search, queries, num_workers=10, timeout=5):
    validate(response)

@operation: defined above

@task_generator: an iterable of things to pass through i_queue to workers

@num_workers: number of child processes to run in parallel, default is 1

@timeout: (required) tests must set an explicit timeout in seconds

@class_config: defined above
    '''
    if timeout is None:
        raise Exception('tests must set an explicit timeout')

    task_generator = iter(task_generator)

    manager = multiprocessing.Manager()
    tasks_remaining = manager.Event()
    tasks_remaining.set()

    ## make I/O queues with no size constraint
    i_queue = manager.Queue() # pylint: disable=E1101
    o_queue = manager.Queue() # pylint: disable=E1101

    pool = multiprocessing.Pool(num_workers, maxtasksperchild=1)

    ## prepare to catch control-C
    global run_many_not_killed
    run_many_not_killed = True
    def clean_exit(sig_num, frame):
        global run_many_not_killed
        run_many_not_killed = False
        logger.debug('attempting clean_exit')
        pool.terminate()
        logger.debug('pool terminated')
        pool.join()
        logger.debug('pool joined. Exiting.')
    for sig_num in [SIGTERM, SIGHUP, SIGABRT, SIGALRM]:
        signal(sig_num, clean_exit)

    ## load up the pool
    async_results = []
    for x in range(num_workers):
        async_res = pool.apply_async(
            worker, 
            args=(operation, i_queue, o_queue, tasks_remaining, class_config,
                  profile))
        async_results.append(async_res)

    start_time = time.time()
    while run_many_not_killed and async_results:
        elapsed = time.time() - start_time
        if elapsed > timeout:
            raise Exception('%.1f seconds elapsed' % elapsed)
        for async_res in async_results:
            try:
                async_res.get(0)
            except multiprocessing.TimeoutError:
                pass
            except Exception, exc:
                logger.critical('worker died!', exc_info=True)
                raise
            else:
                logger.debug('worker finished')
                assert async_res.ready()
                async_results.remove(async_res)
        while elapsed < timeout:
            elapsed = time.time() - start_time
            try:
                resp = o_queue.get(block=False)
                yield resp
            except Queue.Empty:
                # logger.info('no responses from o_queue')
                time.sleep(0.2)
                break
        tasks_added = 0
        while o_queue.qsize() < num_workers * 2 and elapsed < timeout \
              and tasks_added < num_workers * 2:
            elapsed = time.time() - start_time
            try:
                task = task_generator.next()
                i_queue.put(task)
                tasks_added += 1
            except StopIteration:
                tasks_remaining.clear()
                break

    pool.close()
    pool.join()
    logger.debug('finished running %d worker processes' % num_workers)


def pass_through(task, o_queue):
    o_queue.put(task)

def sleeper(task, o_queue):
    time.sleep(4)

def test_multiprocessing_harness():
    num_workers = 10
    task_generator = range(100)
    ret_vals = list(run_many(pass_through, task_generator, 
                             num_workers=num_workers, timeout=20))
    assert set(ret_vals) == set(range(100)), set(range(100)) - set(ret_vals)

def test_multiprocessing_harness_timeout():
    num_workers = 10
    task_generator = range(100)
    with pytest.raises(Exception) as exc:
        list(run_many(sleeper, task_generator, 
                      num_workers=num_workers, timeout=2))
    assert 'seconds elapsed' in str(exc)

def test_multiprocessing_harness_control_C():
    num_workers = 10
    task_generator = range(100)
    alarm(1)
    ret_vals = list(run_many(sleeper, task_generator, 
                             num_workers=num_workers, timeout=20))
    assert len(ret_vals) == 0


class random_inserts(object):
    def __init__(self, config):
        try:
            self.client = kvlayer.client()
            self.client.setup_namespace(dict(t1=1))
            self.one_mb = ' ' * 2**20
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise

    def __call__(self, u, o_queue):
        try:
            self.client.put('t1', ((u,), self.one_mb))
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise
        o_queue.put(u)
        logger.debug('put one_mb at %r', u)


class many_gets(object):
    def __init__(self, config):
        self.client = kvlayer.client()
        self.client.setup_namespace(dict(t1=1))

    def __call__(self, u, o_queue):
        try:
            kvs = list(self.client.get('t1', (u,)))
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise exc
        assert len(kvs[0][1]) == 2**20
        o_queue.put((True, u))
        logger.debug('retrievied one_mb at %r', u)


def test_throughput_insert_random(client, num_workers=5, num_inserts=100,
                                  profile=False):
    client.setup_namespace(dict(t1=1))
    
    total_inserts = num_workers * num_inserts
    task_generator = [uuid.uuid4() for x in xrange(total_inserts)]
    start_time = time.time()
    ret_vals = list(run_many(random_inserts, task_generator, 
                             class_config=client._config,
                             num_workers=num_workers,
                             timeout=total_inserts * 5,
                             profile=profile))
    elapsed = time.time() - start_time
    assert len(ret_vals) == total_inserts
    rate = total_inserts / elapsed
    logger.info('%d MB written in %.1f seconds --> '
                '%.1f MB/sec across %d workers for storage_type=%s',
                total_inserts, elapsed, rate, num_workers,
                client._config['storage_type'])

    if client._config['storage_type'] in ['postgres', 'postgrest', 'accumulo',
                                          'cassandra', 'redis']:
        start_time = time.time()
        count = 0
        for (found, u) in run_many(many_gets, ret_vals, 
                                   class_config=client._config,
                                   num_workers=num_workers,
                                   timeout=total_inserts * 5,
                                   profile=profile):
            if not found:
                raise Exception('failed to find %r' % u)
            count += 1
        elapsed = time.time() - start_time
        assert len(ret_vals) == total_inserts
        rate = total_inserts / elapsed
        logger.info('%d MB read in %.1f seconds --> '
                    '%.1f MB/sec across %d workers for storage_type=%s',
                    total_inserts, elapsed, rate, num_workers,
                    client._config['storage_type'])


class indexer(object):
    '''
    rapidly changes an index in t2
    '''
    def __init__(self, config):
        self.client = kvlayer.client()
        self.client.setup_namespace(dict(t1=1, t2=2))

    def __call__(self, u, o_queue):
        try:
            kvs = list(self.client.scan('t2', ((u,), (u,))))
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise exc
        u1, u2 = kvs[0][0]
        assert u == u1

        #... need to populate t2 initially, and then flow records through it from 0 to 1-8, to 9-24, 25-57, 58-121, etc.


class joiner(object):
    def __init__(self, config):
        self.client = kvlayer.client()
        self.client.setup_namespace(dict(t1=1, t2=2))

    def __call__(self, u, o_queue):
        try:
            kvs = list(self.client.scan('t2', ((u,), (u,))))
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise exc
        u1, u2 = kvs[0][0]
        assert u == u1
        try:
            kvs = list(self.client.get('t1', (u2,)))
        except Exception, exc:
            logger.critical('client failed!', exc_info=True)
            raise exc
        assert len(kvs[0][1]) == 2**20
        o_queue.put((True, u2))
        logger.debug('retrievied one_mb at %r', u2)

@pytest.mark.skipif('True')
def test_throughput_join(client):
    '''measure throughput of reading data from t1 by first looking up the
    key in t2, while t2 is under write load
    '''
    client.setup_namespace(dict(t1=1, t2=2))
    
    num_workers = 10
    num_inserts = 100
    total_inserts = num_workers * num_inserts
    data_ids = [uuid.uuid4() for x in xrange(total_inserts)]
    start_time = time.time()
    ret_vals = list(run_many(random_inserts, data_ids, 
                             class_config=client._config,
                             num_workers=num_workers, timeout=total_inserts/2))
    elapsed = time.time() - start_time
    assert len(ret_vals) == total_inserts
    rate = total_inserts / elapsed
    logger.info('%d MB written in %.1f seconds --> %.1f MB/sec using %d parallel workers for storage_type=%s',
                total_inserts, elapsed, rate, num_workers, client._config['storage_type'])

    ## start tool subprocesses that each run a pool
    #writers = multiprocessing.Process(
        #target=run_many,
        #args=(indexer, data_ids, 
              #class_config=client._config,
              #num_workers=num_workers, timeout=total_inserts/2))))

    ### finish writing this test...

def main():
    parser = argparse.ArgumentParser(usage='kvlayer_throughput_test <storage_type> <storage_address1> <storage_address2> ...',
                                     conflict_handler='resolve')
    parser.add_argument('storage_type')
    parser.add_argument('storage_addresses', default=[], action='append')
    parser.add_argument('--num-workers', default=5, type=int)
    parser.add_argument('--num-inserts', default=100, type=int)
    parser.add_argument('--profile', action='store_true')
    parser.add_argument('--verbose', action='store_true')
    modules = [yakonfig, kvlayer]
    args = yakonfig.parse_args(parser, modules)

    level = logging.INFO
    if args.verbose:
        level = logging.DEBUG
    logging.basicConfig(level=level)

    if not args.storage_addresses:
        sys.exit('must specify at least one storage_address')

    client = kvlayer.client()
    client.delete_namespace()
    test_throughput_insert_random(
        client, 
        num_workers=args.num_workers, num_inserts=args.num_inserts,
        profile=args.profile,
    )
    client.delete_namespace()

if __name__ == '__main__':
    main()



'''Split S3/kvlayer backend for kvlayer.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

'''
from __future__ import absolute_import
import hashlib
import time

import boto

import kvlayer
from kvlayer._abstract_storage import AbstractStorage
from kvlayer._exceptions import ConfigurationError
from kvlayer._utils import join_key_fragments

class SplitS3Storage(AbstractStorage):
    config_name = 'split_s3'
    default_config = {
        'path_prefix': '',
        'kvlayer_prefix': True,
        'retries': 5,
        'retry_interval': 0.1,
    }

    def __init__(self, *args, **kwargs):
        super(SplitS3Storage, self).__init__(*args, **kwargs)

        # Find some credentials
        aws_access_key_id = self._value_or_path('aws_access_key_id')
        aws_secret_access_key = self._value_or_path('aws_secret_access_key')

        # Other things we need to know
        bucket_name = self._config.get('bucket', None)
        if not bucket_name:
            raise ConfigurationError('split_s3 storage requires bucket')
        self.tables = self._config.get('tables', None)
        if not self.tables:
            raise ConfigurationError('split_s3 storage requires tables')
        self.prefix = self._config.get('path_prefix', '')
        if self._config.get('kvlayer_prefix', True):
            self.prefix += '{}/{}/'.format(self._app_name, self._namespace)
        self.retries = self._config.get('retries', 5)
        self.retry_interval = self._config.get('retry_interval', 0.1)

        # Set up the other backend
        if 'kvlayer' not in self._config:
            raise ConfigurationError('split_s3 storage requires '
                                     'second kvlayer configuration')
        self.kvlclient = kvlayer.client(config=self._config['kvlayer'],
                                        app_name=self._app_name,
                                        namespace=self._namespace)

        # Actually connect to S3
        connection = boto.connect_s3(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            # Any sort of connection pooling apparently fails for
            # HTTPS; see https://github.com/boto/boto/issues/1934
            is_secure=False,
        )
        self.bucket = connection.get_bucket(bucket_name)

    def _value_or_path(self, k):
        if k in self._config:
            return self._config[k]
        if k + '_path' in self._config:
            with open(self._config[k + '_path'], 'r') as f:
                return f.read().strip()
        raise ConfigurationError('split_s3 storage requires ' + k)

    def _s3_key(self, table_name, k):
        key = join_key_fragments(k, key_spec=self._table_names[table_name])
        hasher = hashlib.sha256()
        hasher.update(key)
        key_hash = hasher.hexdigest()
        s3_key = '{}{}/{}/{}/{}'.format(self.prefix, table_name,
                                        key_hash[0:2], key_hash[2:4],
                                        key_hash[4:])
        return self.bucket.new_key(s3_key)
            
    def setup_namespace(self, table_names):
        super(SplitS3Storage, self).setup_namespace(table_names)
        self.kvlclient.setup_namespace(table_names)

    def delete_namespace(self):
        '''Deletes all data from all known tables.

        This iterates through known tables and individually deletes
        S3 objects, which may take some time.  It also deletes all data
        from the underlying kvlayer storage.

        '''
        for t in self.tables:
            if t in self._table_names:
                self._clear_table(t)
        self.kvlclient.delete_namespace()

    def clear_table(self, table_name):
        '''Deletes all data from a single table.

        This deletes both the underlying table and the S3 storage for
        every individual object in the table, if the table is backed by
        S3 storage.

        '''
        if table_name in self.tables:
            self._clear_table(table_name)
        self.kvlclient.clear_table(table_name)

    def _clear_table(self, table_name):
        '''Delete every known S3 object for a table.

        This scans the underlying kvlayer table to get keys, and issues
        S3 delete requests for them.

        '''
        for k in self.kvlclient.scan_keys(table_name):
            self._s3_key(table_name, k).delete()

    def put(self, table_name, *keys_and_values, **kwargs):
        '''Store objects in the underlying storage and maybe S3.'''
        if table_name in self.tables:
            # Hybrid S3/kvlayer object.  If there are multiple
            # keys_and_values, and one late in the list fails, the only
            # sane way to do this is to handle them one at a time.
            for (k, v) in keys_and_values:
                self.kvlclient.put(table_name, (k, ''), **kwargs)
                # This didn't raise an exception, so we can put the real
                # value into S3
                self._s3_key(table_name, k).set_contents_from_string(v)
        else:
            # Flat kvlayer object.
            self.kvlclient.put(table_name, *keys_and_values, **kwargs)

    def _get(self, table_name, k):
        '''Get a single object value from S3.

        This implements the retry logic.

        '''
        tries_left = self.retries + 1
        key = self._s3_key(table_name, k)
        while tries_left > 0:
            try:
                return key.get_contents_as_string()
            # Is there something more specific we can catch?
            # Boto has its own retry facility, though that's probably
            # more about failure to connect than failure to retrieve.
            except Exception, e:
                if tries_left == 0:
                    raise
                pass
            tries_left -= 1
            time.sleep(self.retry_interval)

    def scan(self, table_name, *key_ranges, **kwargs):
        '''Combination key/value scan.

        If you only need the keys, :meth:`scan_keys` won't access S3.

        '''
        if table_name in self.tables:
            # Hybrid S3/kvlayer object.  Ask kvlayer for the keys,
            # then S3 for the values.
            for k in self.kvlclient.scan_keys(table_name, *key_ranges,
                                              **kwargs):
                yield (k, self._get(table_name, k))
        else:
            # Flat kvlayer object.
            for kv in self.kvlclient.scan(table_name, *key_ranges, **kwargs):
                yield kv

    def scan_keys(self, table_name, *key_ranges, **kwargs):
        '''Key-only scan.

        This is always a pass-through to the underlying kvlayer and
        never accesses S3.

        '''
        return self.kvlclient.scan_keys(table_name, *key_ranges, **kwargs)

    def get(self, table_name, *keys, **kwargs):
        '''Get specific keys.'''
        for (k,v0) in self.kvlclient.get(table_name, *keys, **kwargs):
            if v0 is not None and table_name in self.tables:
                v = self._get(table_name, k)
            else: # either not present or uninteresting table
                v = v0
            yield (k,v)

    def delete(self, table_name, *keys, **kwargs):
        '''Delete specific keys.'''
        if table_name in self.tables:
            for k in keys:
                self._s3_key(table_name, k).delete()
        self.kvlclient.delete(table_name, *keys, **kwargs)

    def close(self):
        '''Shut down.'''
        self.kvlclient.close()



from __future__ import absolute_import
import uuid
from kvlayer._file_storage import FileStorage

def test_persistence(tmpdir, namespace_string):
    filename = str(tmpdir.join('original'))

    ## Test that we can persist data to a file
    storage = FileStorage(config={'filename': filename},
                          app_name='kvlayer', namespace=namespace_string)
    storage.setup_namespace({'table1': 4, 'table2': 1})

    storage.put('table1', *[((uuid.uuid4(), uuid.uuid4(),
                              uuid.uuid4(), uuid.uuid4()), b'test_data')])
    results = list(storage.scan('table1'))
    assert len(results) == 1
    assert results[0][1] == 'test_data'

    ## Test that we can get same data to from file
    storage = FileStorage(config={'filename': filename},
                          app_name='kvlayer', namespace=namespace_string)
    storage.setup_namespace({'table1': 4, 'table2': 1})
    results2 = list(storage.scan('table1'))
    assert len(results2) == 1
    assert results2[0][1] == 'test_data'

    ## Test that we can get can copy original tables to a new file
    copy_filename = str(tmpdir.join('new'))
    storage = FileStorage(config={'filename': filename,
                                  'copy_to_filename': copy_filename},
                          app_name='kvlayer', namespace=namespace_string)
    storage.setup_namespace({'table1': 4, 'table2': 1})
    results3 = list(storage.scan('table1'))
    assert len(results3) == 1
    assert results3[0][1] == 'test_data'

    ## Assert that all the results are the same
    assert results == results2 == results3
